CLIENT_ID = 'Iv1.b507a08c87ecfe98'

HEADERS = {
    'accept': 'application/json',
    'editor-version': 'Neovim/0.9.5',
    'editor-plugin-version': 'CopilotChat.nvim/2.0.0',
    'content-type': 'application/json',
    'user-agent': 'CopilotChat.nvim/2.0.0',
    # 'accept-encoding': 'gzip,deflate,br'
    }

COPILOT_CHAT_HEADERS = {
        'copilot-integration-id': 'vscode-chat',
        'openai-organization':    'github-copilot',
        'openai-intent':          'conversation-panel'}

TOKEN_FILE = '.copilot-token'

## create a new device code

GITHUB_DEVICE_CODE_URL = 'https://github.com/login/device/code'
DEVICE_CODE_DATA_TEMPLATE = '{{"client_id":"{}","scope":"read:user"}}'

## waiting for a user to authenticate

GITHUB_ACCESS_TOKEN_URL = 'https://github.com/login/oauth/access_token'
ACCESS_TOKEN_DATA_TEMPLATE = '{{"client_id":"{}","device_code":"{}","grant_type":"urn:ietf:params:oauth:grant-type:device_code"}}'

WAIT_AUTH_MAX_RETRIES = 6
WAIT_AUTH_RETRY_INTERVAL = 8

GITHUB_COPILOT_TOKEN_URL = 'https://api.github.com/copilot_internal/v2/token'
GITHUB_COPILOT_CHAT_COMPLETIONS_URL = 'https://api.githubcopilot.com/chat/completions'



import http.server
import requests
import json
import time
import sys
import os

from .config import (
    GITHUB_DEVICE_CODE_URL,
    GITHUB_ACCESS_TOKEN_URL,
    GITHUB_COPILOT_TOKEN_URL,
    HEADERS,
    CLIENT_ID,
    DEVICE_CODE_DATA_TEMPLATE,
    ACCESS_TOKEN_DATA_TEMPLATE,
    WAIT_AUTH_MAX_RETRIES as MAX_RETRIES,
    WAIT_AUTH_RETRY_INTERVAL as RETRY_INTERVAL,
    TOKEN_FILE
)

def is_token_expired(token):

    pairs = token.split(';')
    token_dict = {}

    for pair in pairs:
        key, value = pair.split('=')
        token_dict[key] = value

    if 'exp' in token_dict:
        expiration_timestamp = int(token_dict['exp'])

        current_timestamp = int(time.time())

        if current_timestamp >= expiration_timestamp:
            return True
        else:
            return False
    else:
        return True


def read_token_from_file(token_file):

    if os.path.isfile(token_file):
        try:
            with open(token_file, 'r') as file:
                token = file.read().strip()
                return token
        except IOError as e:
            print(f"could not read token file: {e}")
            return None
    else:
        print(f"don't see a token file: {token_file}")
        return None

def wait_for_access_token(device_code):

    retry_count = 0

    while retry_count < MAX_RETRIES:

        time.sleep(RETRY_INTERVAL)
        retry_count += 1

        resp = requests.post(GITHUB_ACCESS_TOKEN_URL,
                             headers = HEADERS,
                             data = ACCESS_TOKEN_DATA_TEMPLATE.format(CLIENT_ID,
                                                                      device_code))

        parsed = resp.json()

        if 'access_token' in parsed:
            return parsed['access_token']

        elif 'error' in parsed:

            error_message = parsed['error']

            if error_message == 'authorization_pending':
                print('waiting for user authorization...')
            elif error_message == 'slow_down':
                print('rate limit exceeded. waiting for a minute before retrying...')
                time.sleep(60)
            else:
                print(f'error: {error_message}')
                break
        else:
            print('unexpected response format')
            break

    print('maximum retries exceeded. failed to obtain access token.')
    return None

# create github copilot access token
# record the token in a file

def create_access_token(token_file):

    resp = requests.post(GITHUB_DEVICE_CODE_URL,
                         headers = HEADERS,
                         data = DEVICE_CODE_DATA_TEMPLATE.format(CLIENT_ID))

    parsed = resp.json()
    device_code = parsed.get('device_code')
    user_code = parsed.get('user_code')
    verification_uri = parsed.get('verification_uri')

    print(f'browse to {verification_uri} and enter this code "{user_code}" to authenticate')

    access_token = wait_for_access_token(device_code)

    with open(token_file, 'w') as f:
        f.write(access_token)

    return access_token

def refresh_token(token_file=None):

    if token_file is None:
        token_file = os.path.expanduser(TOKEN_FILE)

    access_token = read_token_from_file(token_file)

    if access_token is None:
        access_token = create_access_token(token_file)

    headers = {**HEADERS,
               'authorization': f'token {access_token}'}

    try:
        response = requests.get(GITHUB_COPILOT_TOKEN_URL,
                                headers = headers)
        response.raise_for_status()

    except requests.exceptions.RequestException as e:
        print(f"could not call copilot token referesh API due to: {e}")
        return None

    try:
        parsed = response.json()
        token = parsed.get('token')

        if token is None:
            print("could not extract the token from the API response")
            return None

        return token

    except json.JSONDecodeError:
        print("could not parse the API response as JSON")
        return None



import sys
import os
import time
import argparse

import json

import http.server
import socketserver
import requests

from . import shtoken

from .config import (
    HEADERS,
    COPILOT_CHAT_HEADERS,
    GITHUB_COPILOT_CHAT_COMPLETIONS_URL
)

class Copilot():

    def __init__(self):
        self.token = shtoken.refresh_token()

    def ask(self,
            messages,
            intent=True,
            stream=False,
            reply_as_is=False,   # return the response as is, without any post-processing | parsing
            model='gpt-4',
            n=1,
            temperature=0.1,
            top_p=1):

        if shtoken.is_token_expired(self.token):
            print("refreshing the token...")
            self.token = shtoken.refresh_token()

        headers = {**(HEADERS | COPILOT_CHAT_HEADERS),
                   'authorization': f'Bearer {self.token}'}

        ## proxy all the args copilot
        req = locals()
        req.pop('self', None)
        req.pop('reply_as_is', None)

        try:
            resp = requests.post(GITHUB_COPILOT_CHAT_COMPLETIONS_URL,
                                 headers = headers,
                                 json = req)

            if resp.status_code != 200:
                return {'status': resp.status_code,
                        'text': {'reason': resp.reason,
                                 'text': resp.text}}

            try:

                if reply_as_is:
                    return resp.json()

                parsed = resp.json()

                return {'status': 200,
                        'text': {'reply': self.parse_reply(parsed),
                                 'usage': self.parse_usage(parsed)}}

            except (json.JSONDecodeError, KeyError, IndexError) as e:
                problem = f"error parsing JSON: {str(e)} of this response: {resp.text}"
                return {'status': 400,
                        'text': problem + ", response: " + response.text}

        except requests.exceptions.ConnectionError as e:
            problem = f"connection error: {str(e)}"
            return {'status': 400,
                    'text': problem}

    def parse_reply(self, reply):
        return reply.get('choices')[0].get('message').get('content')

    def parse_usage(self, reply):
        return reply.get('usage')

class ColalamoHandler(http.server.SimpleHTTPRequestHandler):

    def __init__(self, *args, **kwargs):
        self.copilot = kwargs.pop('copilot')
        super().__init__(*args, **kwargs)

    def do_POST(self):

        if self.path == '/ask':

            print("asking... sit tight")

            content_length = int(self.headers['Content-Length'])
            post_data = self.rfile.read(content_length)

            try:
                prompt = json.loads(post_data.decode('utf-8'))

                response = self.copilot.ask(**prompt)

            except json.JSONDecodeError:
                response = {
                    "status": "error",
                    "text": "invalid JSON data"
                }
                self.send_response(400)
            else:
                self.send_response(response['status'])

            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(json.dumps(response['text']).encode('utf-8'))

        else:
            self.send_response(404)
            self.send_header("Content-type", "text/plain")
            self.end_headers()
            self.wfile.write(b"not found => I'm not listening to this endpoint")

class Colalamo():

    def __init__(self, port=4242):
        self.port = port

    def start(self):

        handler = lambda *args, **kwargs: ColalamoHandler(*args,
                                                          **kwargs,
                                                          copilot = Copilot())
        with socketserver.TCPServer(("", self.port), handler) as httpd:
            print(f"colalamo is listening on 0.0.0.0:{self.port}")
            print(f"ask away at \"/ask\"")
            print(f"""example: curl http://localhost:{self.port}/ask -X POST -d '{{"messages": [{{"role": "user", "content": "explain how multi-head attention work"}}]}}'""")
            httpd.serve_forever()




def parse_arguments():
    description = """
colalamo is a proxy server that sends custom requests via GitHub Copilot to its backing LLMs

usage: colalamo [--port PORT]
"""
    parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawDescriptionHelpFormatter, usage=argparse.SUPPRESS)

    parser.add_argument("-p", "--port", type=int, default=4242, help="colalamo will listen on this port number (default: 4242)")

    return parser.parse_args()

def main():

    args = parse_arguments()
    port = args.port

    colalamo = Colalamo(port)
    colalamo.start()

    # prompt = "write a user story for a landing page that has three charts with totals, under each chart that is a ui grid with data"
    # copilot = Copilot()

    # answer = copilot.ask({'messages': [{'content': prompt,
    #                                     'role': 'user'}]})

    # print(f"answer: {answer}")

if __name__ == '__main__':
    main()
